#!/usr/bin/env -S blender -P
# Note: the shebang line above requires coreutils 8.30 and above
# It is recommended to invoke this script with facade.py for portability

import sys
import bpy
import argparse
import math
import os
from math import pi as PI
import numpy as np
from scipy.spatial.transform import Rotation, Slerp

def norm(vec):
    return np.linalg.norm(vec)

def normalized(vec):
    a = np.array(vec)
    return a / norm(vec)

def distance(q1, q2, w1, w2):
    dq = abs(np.dot(q1[3:7], q2[3:7]))
    #print(f'dq: {dq}')
    dq = np.arccos(np.clip(dq, a_min=-1.0, a_max=1.0))
    return w1 * ((q1[0] - q2[0])**2 + (q1[1] - q2[1])**2 + (q1[2] - q2[2])**2 )**(1/2) + w2 * dq

def quat_distance(q1,q2):
    return (q1[0]-q2[0])**2 + (q1[1] - q2[1])**2 + (q1[2] - q2[2])**2 + (q1[3] - q2[3])**2

def _mix(tau, key_0, key_1):
    t = (1.0 - tau) * key_0[0] + tau * key_1[0]
    rots = Rotation.from_quat([key_0[1].as_quat(), key_1[1].as_quat()])
    slerp = Slerp([0.0, 1.0], rots)
    r, = slerp([tau])
    return (t, r)

def _add_key(rob, t, quat, frame):
    rob.rotation_mode = 'QUATERNION'
    rob.location.x = t[0]
    rob.location.y = t[1]
    rob.location.z = t[2]
    rob.rotation_quaternion.x = quat[0]
    rob.rotation_quaternion.y = quat[1]
    rob.rotation_quaternion.z = quat[2]
    rob.rotation_quaternion.w = quat[3]
    # rob.select = True
    rob.select_set(True)
    rob.keyframe_insert(data_path="location", frame=frame)
    rob.keyframe_insert(data_path="rotation_quaternion", frame=frame)
    # rob.select = False
    rob.select_set(False)

def parse_args():
    p = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('env', help='ENV geometry')
    p.add_argument('rob', help='ROB geometry')
    # p.add_argument('qpath', help='Vanilla configuration path')
    p.add_argument('sol', help='The solution file (in npz)')
    p.add_argument('--total_frames', help='Total number of frames to render', type=int, default=1440)
    p.add_argument('--use_original_path', help='Use original path instead of simplified path',
                   action='store_true')
    argv = sys.argv
    return p.parse_args(argv[argv.index("--") + 1:])

def main():
    args = parse_args()

    # Load meshes
    bpy.ops.import_scene.obj(filepath=args.env, axis_forward='Y', axis_up='Z')
    env = bpy.context.selected_objects[0]
    env.name = 'Env'
    bpy.ops.object.shade_smooth()

    bpy.ops.import_scene.obj(filepath=args.rob, axis_forward='Y', axis_up='Z')
    rob = bpy.context.selected_objects[0]
    rob.name = 'Rob'
    bpy.ops.object.shade_smooth()

    bpy.data.meshes.remove(bpy.data.meshes['Cube'])
    bpy.data.objects.remove(bpy.data.objects['Light'])

    # matrix detailing path of the object. framesamples[i] is coordinates of the robot at step i
    frameSamples = []
    key_confs = []
    # distance travelled per step, using weighted distance. Determines keyrame placement
    distances = [0]
    rotationWeight = 50     #weight to rotational component of distance
    totalDist = 0           #total distance of path.

    # input file location, change as required
    sol_dic = np.load(args.sol)
    path_key = 'OMPL_PATH' if args.use_original_path else 'SIMPLIFIED_PATH'
    ompl_q = sol_dic[path_key]
    O = sol_dic['REFERENCE_ORIGIN']

    prevQ = []
    for el in ompl_q:
        pass
        el = [float(el) for el in el]
        wlast_quat = el[3:]
        T = el[:3]
        if not prevQ:
            prevQ = wlast_quat
        d1 = quat_distance(prevQ, wlast_quat)
        d2 = quat_distance(prevQ, -np.array(wlast_quat))
        if (d1 <= d2):
            Q = wlast_quat
        else:
            Q = (-np.array(wlast_quat)).tolist()
        res = T
        res.extend(Q)
        if frameSamples:
            d = distance(frameSamples[-1], res, 1, rotationWeight)
            distances.append(distances[-1] + d)
            totalDist += d
        frameSamples.append(res)
        R = Rotation.from_quat(el[3:7])
        key_confs.append((np.array(el[:3]), R))

    desiredFrames = args.total_frames
    bpy.types.PreferencesEdit.keyframe_new_interpolation_type = 'LINEAR'

    distance_per_frame = 1.0 / float(desiredFrames) * totalDist
    rob = bpy.data.objects["Rob"]

    for frame in range(desiredFrames):
        d = frame * distance_per_frame
        index_1 = np.argmax(distances > d)
        # print(f'index_1 {index_1}')
        index_0 = index_1 - 1
        d_0 = distances[index_0]
        d_1 = distances[index_1]
        key_0 = key_confs[index_0]
        key_1 = key_confs[index_1]
        tau = (d - d_0) / (d_1 - d_0)
        t,r = _mix(tau, key_0, key_1)
        t = t - r.apply(O) # Translate back to vanilla
        quat = r.as_quat()
        _add_key(rob, t, quat, frame+1)


if __name__ == "__main__":
    main()
